Require Import VST.floyd.proofauto.
Local Open Scope logic. 

(*Model-level representation of master and clock*)

Definition Time : Type := Z.

Definition ClockID := nat.

(*reliable friends/clocks occasionally give us a call/synchronize - those who don't
  would eventually prevent us from resetting the clock so let's purge them after a timeout of our choice*)

(*I believe the token handed out to the lost clock is not a problem: the clock won't be able
  to sync (since purging sets L i=None), but will be able to switch to (a different, by
  separation conjunction) master. ToDo: verify a (couple of?) clients exercising this*)

(*Question: the timeout is only a semantic concept, namely a WITH-list argument to newMaster, NOT
  an argument of the C function, and is hence not represented in the C data structure of masters. 
  Is that a bug or a feature?
  Also, purging only happens semantically, as (just as in BN) next to no info about the
  clocks is stored in the master. *) 

(*MyTime, last-sync'ed table, timeout*)
Definition MasterRepTp:Type := Time * (ClockID -> option Time) * Time.

Definition emptyMaster: ClockID -> option Time := fun i => None.

Definition upd (T:MasterRepTp) (i:nat):MasterRepTp :=
  match T with (t,L,u) =>
    (t, fun j => if beq_nat i j then Some t else L j, u)
  end.

Definition kill (T:MasterRepTp) (i:nat):MasterRepTp :=
  match T with (t,L,u) =>
    (t, fun j => if beq_nat i j then None else L j,u)
  end.

Definition newMaster u: MasterRepTp:= (0, emptyMaster,u).

Definition ClockRepTp:Type := Time * ClockID * val.

Definition newEntry (T:MasterRepTp) (C:ClockRepTp) (U:MasterRepTp) (m:val):Prop :=
  match T with (t, L,u) =>
    exists i, L i = None /\ U = upd T i /\ C = (t,i,m)
  end.

Definition purge (T:MasterRepTp): MasterRepTp :=
  match T with (t, L,u) =>
    (t, fun i => match L i with
                     None => None 
                   | Some s => if zlt (s+u) t then None else Some s
                 end,u)
  end.

(*Master can hand out a token of friendship (whose type is secret to master).
  But note that neither MToken nor MFriend show up in any ASI specs of 
  either Master nor Clock, or in the type ClockRepTp . Instead, we'll roll
  holding the friendship token into the definition of Crep (so so the
  clock APD instance will be parametric in a Master APD instance), and model
  the handing over of tokens using the entailments Mrep_connect (Barnett-
  Naumann's attach/connect) and Mrep_detach (BN's detach). These entailments
  play the role of BN's verification-only functions.

  Update: the introduction of reconnect_spec lead to 1 occurence of
  MFriend in a spec; exposure of axiom MFriend_emp reestablishes the fact
  the Friends need not be exposed in specs. *)
Record MasterAPD := {
  Mrep: MasterRepTp -> val -> mpred;
  Mrep_ptrnull: forall T m, Mrep T m |-- !!(is_pointer_or_null m);

  (*Timeout positive, and t in signed-representable range*)
  Mrep_range: forall t L u m, Mrep (t,L,u) m |-- !!(0< u /\ 0 <= t <= Int.max_signed);

  MToken: Type;
  (*A frienship token that Master hands to a Clock when the Clock connects,
    modeling the 'attach'; Needs to be presented for detaching*)
  MFriend: ClockID -> MToken -> val -> mpred;

  MFriend_emp: forall i q m, MFriend i q m |-- emp;

  (*alternative: initialize i to 0 in L*)
  Mrep_connect: forall t L u m, Mrep (t,L,u) m |-- 
                EX i q, !!(L i = None) && MFriend i q m * Mrep (upd (t,L,u) i) m;

  Mrep_detach: forall i q T m, (MFriend i q m * Mrep T m)%logic |-- Mrep (kill T i) m;

  Mrep_rely: forall T m, (Mrep T m)%logic |-- Mrep (purge T) m;


  Mrep_friend: forall t L u m, Mrep (t,L,u) m |-- !!(forall i ti, L i = Some ti -> t-u <= ti <= t);

}.

Record ClockAPD := {
  Crep: ClockRepTp -> val -> mpred
}.
