Require Import VST.floyd.proofauto. (*TODO: Narrow this*)

Record PosSubT (S T:Type) := {
  project: S -> T; (*implicit coercion*)
  recomb: S -> T -> S; (*put*)
  Ax1: forall s t, project (recomb s t) = t;
  Ax2: forall s, recomb s (project s) = s;
  Ax3: forall s t1 t2, recomb (recomb s t1) t2 = recomb s t2
}.

Definition PosSubTProp (S T:Type) : Prop :=
  exists (project: S -> T) (recomb: S -> T -> S), 
  (forall s t, project (recomb s t) = t) /\
  (forall s, recomb s (project s) = s) /\
  (forall s t1 t2, recomb (recomb s t1) t2 = recomb s t2).

Section PosFun.
(*Positive function space operator preserves positive subtyping*)
Variable S T U:Type.
Variable subST:PosSubT S T.

Notation p := (project _ _ subST).
Notation r := (recomb _ _ subST).

Definition p_fun: (U->S) -> (U->T) := fun f u => p (f u).
Definition r_fun: (U->S) -> (U->T) -> (U -> S) := fun f g u => r (f u) (g u).
Lemma Ax1_fun s t: p_fun (r_fun s t) = t.
Proof. unfold p_fun, r_fun. extensionality u. apply (Ax1 _ _ subST). Qed.
Lemma Ax2_fun s: r_fun s (p_fun s) = s.
Proof. unfold p_fun, r_fun. extensionality u. apply (Ax2 _ _ subST). Qed.
Lemma Ax3_fun s t1 t2: r_fun (r_fun s t1) t2 = r_fun s t2.
Proof. unfold r_fun. extensionality u. apply (Ax3 _ _ subST). Qed.

Definition PosSubT_PosFun: PosSubT (U -> S) (U -> T) :=
Build_PosSubT _ _ _ _ Ax1_fun Ax2_fun Ax3_fun.

End PosFun.

Section PosProduct.
(*Product operator preserves positive subtyping*)
Variable S1 S2 T1 T2:Type.
Variable subST1:PosSubT S1 T1.
Variable subST2:PosSubT S2 T2.

Notation p1 := (project _ _ subST1).
Notation r1 := (recomb _ _ subST1).
Notation p2 := (project _ _ subST2).
Notation r2 := (recomb _ _ subST2).

Definition p_prod: (S1 * S2) -> (T1 * T2) := fun s => (p1 (fst s), p2 (snd s)).
Definition r_prod: (S1 * S2) -> (T1 * T2) -> (S1 * S2) := 
            fun s t => (r1 (fst s) (fst t), r2 (snd s) (snd t)).
Lemma Ax1_prod s t: p_prod (r_prod s t) = t.
Proof. unfold p_prod, r_prod. simpl.
  rewrite (Ax1 _ _ subST1), (Ax1 _ _ subST2).
  destruct t; trivial. 
Qed.
Lemma Ax2_prod s: r_prod s (p_prod s) = s.
Proof. unfold p_prod, r_prod. simpl.
  rewrite (Ax2 _ _ subST1), (Ax2 _ _ subST2).
  destruct s; trivial.
Qed.
Lemma Ax3_prod s t1 t2: r_prod (r_prod s t1) t2 = r_prod s t2.
Proof. unfold r_prod. simpl.
  rewrite (Ax3 _ _ subST1), (Ax3 _ _ subST2).
  destruct s; trivial.
Qed.

Definition PosSubT_Prod: PosSubT (S1 * S2) (T1 * T2) :=
Build_PosSubT _ _ _ _ Ax1_prod Ax2_prod Ax3_prod.

End PosProduct.

Section PosRefl.
Variable U:Type.
Definition p_refl: U -> U := id.
Definition r_refl (u v:U): U := v. 
Lemma Ax1_refl s t: p_refl (r_refl s t) = t.
Proof. reflexivity. Qed.
Lemma Ax2_refl s: r_refl s (p_refl s) = s.
Proof. reflexivity. Qed.
Lemma Ax3_refl s t1 t2: r_refl (r_refl s t1) t2 = r_refl s t2.
Proof. reflexivity. Qed.

Definition PosSubT_Refl: PosSubT U U :=
Build_PosSubT _ _ _ _ Ax1_refl Ax2_refl Ax3_refl.
End PosRefl.

Section PosTrans.
Variable S R T:Type.
Variable subSR:PosSubT S R.
Variable subRT:PosSubT R T.

Notation psr := (project _ _ subSR).
Notation rsr := (recomb _ _ subSR).
Notation prt := (project _ _ subRT).
Notation rrt := (recomb _ _ subRT).

Definition p_trans: S -> T := fun s => prt (psr s).
Definition r_trans: S -> T -> S := fun s t => rsr s (rrt (psr s) t).

Lemma Ax1_trans s t: p_trans (r_trans s t) = t.
Proof. unfold p_trans, r_trans. rewrite 2 Ax1; trivial. Qed.

Lemma Ax2_trans s: r_trans s (p_trans s) = s.
Proof. unfold p_trans, r_trans.
specialize (Ax2 _ _ subSR s); intros.
rewrite <- (Ax2 _ _ subRT (psr s)) in H; trivial.
Qed.

Lemma Ax3_trans s t1 t2: r_trans (r_trans s t1) t2 = r_trans s t2.
Proof. unfold p_trans, r_trans. rewrite Ax3, Ax1, Ax3. trivial. Qed.

Definition PosSubT_Trans: PosSubT S T :=
Build_PosSubT _ _ _ _ Ax1_trans Ax2_trans Ax3_trans.
End PosTrans.

Section PosTop.
Variable S:Type.
Definition p_top: S -> unit := fun _ => tt.
Definition r_top: S -> unit -> S := fun s _ => s.

Lemma Ax1_top s t: p_top (r_top s t) = t.
Proof. destruct t. reflexivity. Qed.

Lemma Ax2_top s: r_top s (p_top s) = s.
Proof. reflexivity. Qed.

Lemma Ax3_top s t1 t2: r_top (r_top s t1) t2 = r_top s t2.
Proof. reflexivity. Qed.
Definition PosSubT_Top: PosSubT S unit :=
Build_PosSubT _ _ _ _ Ax1_top Ax2_top Ax3_top.
End PosTop.

Section PosAll.
Variable U:Type.
Variable S T: Type -> Type. 
Variable Hyp: forall X, (PosSubT X U -> PosSubT (S X) (T X)). 

Definition p_all: (forall X : Type, PosSubT X U -> S X) ->
                    forall X : Type, PosSubT X U -> T X :=
fun HS Y SUBY => (project _ _ (Hyp Y SUBY) (HS _ SUBY)).

Definition r_all: (forall X : Type, PosSubT X U -> S X) ->
                   (forall X : Type, PosSubT X U -> T X) ->
                   forall X : Type, PosSubT X U -> S X :=
fun HS HT Y SUBY => (recomb _ _ (Hyp Y SUBY) (HS _ SUBY) (HT _ SUBY)).

Lemma Ax1_all s t: p_all (r_all s t) = t.
Proof. unfold p_all, r_all. extensionality Y HY. rewrite Ax1; trivial. Qed.

Lemma Ax2_all s: r_all s (p_all s) = s.
Proof. unfold p_all, r_all. extensionality Y HY. rewrite Ax2; trivial. Qed.

Lemma Ax3_all s t1 t2: r_all (r_all s t1) t2 = r_all s t2.
Proof. unfold p_all, r_all. extensionality Y HY. rewrite Ax3; trivial. Qed.

Definition PosSubT_All: PosSubT (forall X, PosSubT X U -> S X) 
                                (forall X, PosSubT X U -> T X):=
Build_PosSubT _ _ _ _ Ax1_all Ax2_all Ax3_all.
End PosAll.

Section Section5.
(*
Fixpoint NestedTuple (Ts: list Type): Type  :=
  match Ts with
    nil => unit
  | T1::Ts' => T1 * NestedTuple Ts'
  end.
*)
(*Goal (NestedTuple [Z;Z]) = unit.*)
(*So maybe encode records directly*)
(*
Section PosList.
Variable S T:list Type.
Check (list Type).

Definition PosSubT_All (U W:list Type): PosSubT U W.
Variable T: Type -> Type. 
Variable Hyp: forall X, (PosSubT X U) -> PosSubT (S X) (T X). 

Definition p_all: (forall X : Type, PosSubT X U -> S X) ->
                    forall X : Type, PosSubT X U -> T X :=
fun HS Y SUBY => (project _ _ (Hyp Y SUBY) (HS _ SUBY)).

Definition r_all: (forall X : Type, PosSubT X U -> S X) ->
                   (forall X : Type, PosSubT X U -> T X) ->
                   forall X : Type, PosSubT X U -> S X :=
fun HS HT Y SUBY => (recomb _ _ (Hyp Y SUBY) (HS _ SUBY) (HT _ SUBY)).

Lemma Ax1_all s t: p_all (r_all s t) = t.
Proof. unfold p_all, r_all. extensionality Y HY. rewrite Ax1; trivial. Qed.

Lemma Ax2_all s: r_all s (p_all s) = s.
Proof. unfold p_all, r_all. extensionality Y HY. rewrite Ax2; trivial. Qed.

Lemma Ax3_all s t1 t2: r_all (r_all s t1) t2 = r_all s t2.
Proof. unfold p_all, r_all. extensionality Y HY. rewrite Ax3; trivial. Qed.

Definition PosSubT_All: PosSubT (forall X, (PosSubT X U) -> S X) 
                                  (forall X, (PosSubT X U) -> T X):=
Build_PosSubT _ _ _ _ Ax1_all Ax2_all Ax3_all.
End PosAll.*)


End Section5.